
;place the PROTO near beginning of file:

uArbitrary2Str  PROTO   :DWORD,:DWORD,:DWORD,:DWORD

;place the PROC in the .code segment:

;-------------------------------------------------------------------------------------

        OPTION  EPILOGUE:NONE
        OPTION  PROLOGUE:NONE

uArbitrary2Str  PROC lpBinary:DWORD,dwNumberOfDWords:DWORD,lpBuffer:DWORD,dwSize:DWORD
;
; CONVERT UNSIGNED ARBITRARY INTEGER TO ASCII DECIMAL STRING
;
;Original 16-bit code by Richard Pavlicek, 1998
;Modified for 32-bit and features added by DednDave, 2009 (revision 5)
;
;  Besides 32-bit windows compatibility, one of the most signifigant improvements made
;is that the original binary input value is not destroyed. Fairly comprehensive error
;checking has been added. Several speed improvements have been made, although the
;routine is certainly still not optimal. It is accurate and reliable. It can handle
;very large input values, limited only by available addressing space.
;
;-----------------------------------------
;Usage:
;
;       INVOKE  uArbitrary2Str,
;               lpBinary,
;               dwNumberOfDWords,
;               lpBuffer,
;               dwSize
;
;         lpBinary = Offset of unsigned binary input value
; dwNumberOfDWords = Number of dwords in input value
;         lpBuffer = Offset of output buffer
;           dwSize = Size of the output buffer in bytes
;
;-----------------------------------------
;Returns:
;
; edx = Pointer to first byte of output string
; ecx = Length of output string, in bytes
; eax = Return status:
;   eax = 0   ERROR_SUCCESS
;             "The operation completed successfully."
;          a) no errors:
;               the ASCII decimal string pointer and length are returned
;   eax = 8   ERROR_NOT_ENOUGH_MEMORY
;             "Not enough storage is available to process this command."
;          a) the input value will not fit on the stack with 512 bytes remaining:
;               the specified output buffer pointer with length = 0 is returned
;   eax = 87  ERROR_INVALID_PARAMETER
;             "The parameter is incorrect."
;          a) dwNumberOfDWords is 0 or negative:
;               the specified output buffer pointer with length = 0 is returned
;          b) dwSize is 0 or negative:
;               the specified output buffer pointer with length = 0 is returned
;   eax = 122 ERROR_INSUFFICIENT_BUFFER
;             "The data area passed to a system call is too small."
;          a) the output buffer is too small:
;               the specified output buffer pointer with length = 0 is returned
;         (windows list of error codes and descriptions was poorly devised)
;
;-----------------------------------------
;Notes:
;
; 1) The output string is right-justified in the output buffer field.
; 2) No verification is made to insure the address parameters are valid (i.e. access
;    privileges). However, if adding the output buffer address to its' size causes
;    overflow, the size is flagged as invalid. Similarly, if adding the input address
;    to the input length causes overflow, the length is flagged as invalid. The later
;    may cause confusion, as one might expect "ERROR_NOT_ENOUGH_MEMORY", instead.
; 3) The maximum input length is dependant on available stack space. The binary input
;    value must fit on the stack, with 512 bytes remaining. Of course, the resulting
;    string must also fit into the output buffer. A 32-dword input value may result
;    in a 309-byte output string. The routine fills 8 bytes of the output buffer in
;    each loop pass. So, when processing a 32-dword input value, a 312-byte buffer
;    may be required. Generally speaking, if the output buffer is 2.5 times the size
;    of the input value, it should be large enough for the resulting string.
;    output buffer in bytes = 10 x input value length in dwords
;    This rule of thumb works except for 1-dword, 2-dword, and 3-dword input values,
;    which require 16, 24, and 32 byte buffers, respectively. That means that the
;    theoretical maximum input size that may be accomodated is roughly
;    (100000000h-200h)/4.5 bytes, or 238,609,265 dwords. Much of the 4 GB address
;    space is not usable, so the actual maximum input size will be much smaller. A
;    further limitation is made by the assemblers, which may not allow definition of
;    arrays so large. This limitation varies with the assembler used. Input values
;    and output buffers are not neccessarily created in this manner.
;
;-----------------------------------------

;set up the stack frame

        push    ebx               ;save registers
        push    esi
        push    edi
        push    ebp
        mov     ebp,esp           ;stack frame pointer

;verify valid input length parameter

        mov     ecx,[ebp+24]      ;parm2 = input length in dwords
        xor     eax,eax           ;multi-purpose zero
        cmp     eax,ecx           ;error if length < +1
        jge     A2Str6            ;error - input length invalid

;reduce input length for any high-order zero dwords

;eax = 0
;ecx = input length in dwords

        mov     edi,ecx           ;dword count in edi
        dec     edi               ;point to last dword
        shl     edi,2             ;in bytes
        jc      A2Str7            ;error - input length too large

        add     edi,[ebp+20]      ;parm1 = input value address
        jc      A2Str6            ;error - input length invalid

        std                       ;scan downward
        repz    scasd             ;skip over the 0's
        inc     ecx               ;include the last non-zero dword

;verify sufficient stack space

;df  = std
;eax = 0
;ecx = adjusted input length in dwords
;edi = offset of second-to-last dword of input value

        mov     edx,ecx           ;adjusted dword count now
        lea     ebx,[esp-200h]    ;stack with 512 byte reserve
        shl     edx,2             ;adjusted byte count
        cmp     ebx,edx           ;stack overflow ?
        jb      A2Str7            ;error - input length too large

;verify valid output buffer size parameter

;df  = std
;eax = 0
;ebx = esp-200h
;ecx = adjusted input length in dwords
;edx = adjusted input length in bytes
;edi = offset of second-to-last dword of input value

        mov     esi,[ebp+32]      ;parm4 = size of output buffer in bytes
        cmp     eax,esi           ;error if size < +1
        jge     A2Str6            ;error - invalid output buffer size

        add     esi,[ebp+28]      ;parm3 = address of output buffer
        jc      A2Str6            ;error - invalid output buffer size

        mov     [ebp+32],esi      ;replace length with end address

;probe the stack down to the required address if needed

;df  = std
;eax = 0
;ebx = esp-200h
;ecx = adjusted input length in dwords
;edx = adjusted input length in bytes
;edi = offset of second-to-last dword of input value

        mov     esi,800h          ;2048 byte half-page size
        cmp     edx,esi           ;need to probe stack ?
        jbe     A2Str1            ;no - skip it

        neg     esi               ;minus makes a mask
        sub     ebx,edx           ;target stack address
        and     esp,esi           ;clear low bits of esp

A2Str0: mov     [esp+esi],eax     ;probe the page
        add     esp,esi           ;next half-page
        cmp     esp,ebx           ;target reached ?
        ja      A2Str0            ;no - until it is

        mov     esp,ebp           ;yes - restore esp

A2Str1: sub     esp,edx           ;make room for value on stack

;copy the input value onto the stack

;df  = std
;eax = 0
;ecx = adjusted input length in dwords
;edi = offset of second-to-last dword of input value

        lea     esi,[edi+4]       ;last dword of input value
        lea     edi,[ebp-4]       ;last dword of destination
        mov     ebx,ecx           ;save dword count in ebx
        mov     edx,edi           ;save the working pointer
        rep     movsd             ;copy the input value
        xchg    eax,edx           ;0 in edx, pointer in eax
        mov     edi,[ebp+32]      ;end of output string
        xchg    eax,esi           ;pointer in esi
        cld                       ;direction = up

;outer divide/convert loop

;ebx = adjusted input length in dwords
;edx = 0
;esi = pointer to high order input dword (stack copy)
;edi = pointer to byte after end of input buffer

        ALIGN   4                 ;align the outer loop

A2Str2: nop                       ;alignment
        mov     [ebp+24],ebx      ;save dword count
        nop                       ;alignment
        mov     [ebp+20],esi      ;save high pointer

A2Str3: nop                       ;alignment
        sub     edi,8             ;make room for 8 decimal digits
        nop                       ;alignment
        cmp     edi,[ebp+28]      ;test for buffer overflow
        nop                       ;align the inner loop
        jb      A2Str8            ;error - output buffer overflow

;inner loop divides entire input value by 100 million

        mov     ecx,100000000     ;100 million to get 8 digits

A2Str4: mov     eax,[esi]         ;get dword of dividend
        div     ecx               ;divide by 100 million
        mov     [esi],eax         ;save quotient dword
        sub     esi,4             ;next dword
        dec     ebx               ;done ?
        jnz     A2Str4            ;no - repeat

;continuation of outer loop: edx = binary value for the low 8 decimal digits

        xchg    eax,ebx           ;zero in eax
        mov     ecx,10000         ;divisor to make 4-digit groups
        xchg    eax,edx           ;remainder (0-99999999)
        mov     esi,30303030h     ;ascii numchar bits in esi
        div     ecx               ;dx = 0-9999, ax = 0-9999
        mov     cl,100            ;divisor to make 2-digit groups
        div     cl                ;ah = 0-99, al = 0-99
        mov     ebx,eax           ;save al
        mov     al,ah             ;least significant 2 digits
        aam                       ;ah = 0-9, al = 0-9
        bswap   eax               ;reverse and save in upper half of eax
        mov     al,bl             ;recall al
        aam                       ;ah = 0-9, al = 0-9
        xchg    al,ah             ;reverse order
        or      eax,esi           ;make ascii
        mov     [edi],eax         ;most significant 4 digits
        xchg    eax,edx           ;least significant (0-9999)
        div     cl                ;ah = 0-99, al = 0-99
        mov     ebx,eax           ;save al
        mov     al,ah             ;least significant 2 digits
        aam                       ;ah = 0-9, al = 0-9
        bswap   eax               ;reverse and save in upper half of eax
        mov     al,bl             ;recall al
        aam                       ;ah = 0-9, al = 0-9
        xchg    al,ah             ;reverse order
        mov     ebx,[ebp+24]      ;recall dword count
        or      eax,esi           ;make ascii
        xor     edx,edx           ;dual-purpose zero
        mov     esi,[ebp+20]      ;recall high dword pointer
        mov     [edi+4],eax       ;least significant 4 digits
        cmp     edx,[esi]         ;is high dword zero yet?
        jnz     A2Str3            ;no, continue with same count

        sub     esi,4             ;yes, point to next dword
        dec     ebx               ;reduce dword count
        jnz     A2Str2            ;until done

;suppress leading 0(s)

        mov     al,30h            ;ascii '0'
        mov     ecx,8             ;maximum needed to check
        repz    scasb             ;advance edi past first non-zero digit
        dec     edi               ;point to first non-zero digit in string

;set return values

        xchg    eax,ebx           ;eax = 0 if no error
        mov     ecx,[ebp+32]      ;recall string tail pointer
        mov     edx,edi           ;edx = offset of ascii string
        sub     ecx,edi           ;ecx = number of ascii digits

;restore registers and exit

A2Str5: leave                     ;mov esp,ebp - pop ebp
        pop     edi               ;restore registers
        pop     esi
        pop     ebx
        ret     16                ;return and discard 4 dwords

;invalid input length parameter or invalid output size parameter

A2Str6: mov     eax,87            ;return code in eax
        jmp short A2Str9          ;error exit

;input value too large (not enough memory)

A2Str7: mov     eax,8             ;return code in eax
        jmp short A2Str9          ;error exit

;output buffer overflow

A2Str8: mov     eax,122           ;return code in eax

;error exit

A2Str9: cld                       ;insure direction = up
        xor     ecx,ecx           ;length = 0
        mov edx,[ebp+28]          ;output buffer address
        jmp     A2Str5            ;exit the routine

uArbitrary2Str  ENDP

        OPTION  EPILOGUE:EpilogueDef
        OPTION  PROLOGUE:PrologueDef

;-------------------------------------------------------------------------------------
